/****
* 소스 사용에는 어떠한 제한도 없습니다.
* 본 소스를 사용하여 발생한 내용에 대해서는 책임을 지지 않습니다. 
* 버그 수정이나 업그레이는 홈페이지(http://www.cocoadev.co.kr/)에서 확인해 주세요.
* 마지막 수정: 2007-12-14 by zzerr
**/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/socket.h>
#include <sys/errno.h>
#include <errno.h>

#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "brss.h"

/** xml tags */
char* CBlogRss::m_paszTag[MAX_PTAG] = { 
     "title",
     "link",
     "description",
     "pubDate",
	 "item"
};

CBlogRss::CBlogRss()
{
	m_bHeadSet = false;
	m_nSocket = -1;
	m_nPostCount = 0;

	m_nErrorLine = 0;
}

CBlogRss::~CBlogRss()
{
	for(int i = 0; i < m_nPostCount; i++)
	{
		if(m_tBlogPosts[i]->pszDescription)
			delete [] m_tBlogPosts[i]->pszDescription;

		delete m_tBlogPosts[i];
	}
}

T_BlogPost* CBlogRss::GetBlogPost(int nPos)
{
	if(nPos < 0 || nPos >= m_nPostCount)
		return 0;

	return m_tBlogPosts[nPos];
}

bool CBlogRss::GetData(const char* pszUrl, const char* pszPath)
{
	if(Connect(pszUrl, pszPath) == false)
	{
		SetErrorLine(__LINE__);
		return false;
	}

	unsigned int nRt, nLen;
    char aszTemp[256];    

	/** Request 데이터를 전송 */
	sprintf(aszTemp, "GET %s HTTP/1.0\r\nHost: %s\r\n", pszPath, pszUrl);    
	strcat(aszTemp, "User-Agent: Mozilla/5.0 (Macintosh; U; PPC Mac OS X Mach-O; ko; rv:1.8.1.9) Gecko/20071025 Firefox/2.0.0.9\r\n");
    strcat(aszTemp, "Content-Type: application/x-www-form-urlencoded\r\n");
    strcat(aszTemp, "Content-Length: 0\r\n\r\n");

    nLen = strlen(aszTemp);

    nRt = Write(aszTemp, nLen);
	if(nRt != nLen)
	{
		SetErrorLine(__LINE__);
		return false;
	}

	/** RSS 데이터 수신 */
	char* pszBuff = new char[RBUFFER_SIZE];
	if(!pszBuff)
	{
		SetErrorLine(__LINE__);
		return false;
	}

    if(Read(pszBuff, RBUFFER_SIZE) < 10)
	{
		SetErrorLine(__LINE__);
		delete [] pszBuff;
		return false;
	}

	close(m_nSocket);
	m_nSocket = -1;

    char* pTemp = strstr(pszBuff, "<channel>");
    if(!pTemp)
    {    
		SetErrorLine(__LINE__);
		delete [] pszBuff;
        return false;
    }

	/** 수신 데이터 파싱 */
	ParseData(pTemp);
	delete [] pszBuff;

    return true;
}

bool CBlogRss::AddPost()
{
	if(m_nPostCount >= MAX_READ_COUNT)
	{
		SetErrorLine(__LINE__);
		return false;
	}

	m_tBlogPosts[m_nPostCount] = new T_BlogPost;
	memset(m_tBlogPosts[m_nPostCount], 0, sizeof(T_BlogPost));

	m_nPostCount++; 

	return true;
}

bool CBlogRss::SetPostData(int nType, const char* pszValue)
{
	int nPos = m_nPostCount - 1;
	int nLen = strlen(pszValue);
	int nCopyLen = nLen;

	/* 텍스트 길이 검사(포스트 내용일 경우에는 예외) */
	if(nLen >= TEXT_SIZE) 
	{
		nCopyLen = TEXT_SIZE - 1;
	}

	/* m_bHeadSet이 설정이 안되어 있으면 블로그 정보에 되어 있으면 포스트 정보에 기록 */
	switch(nType)
	{
		case PTAG_TITLE :
			if(m_bHeadSet == false)
			{
				strncpy(m_tBlog.aszTitle, pszValue, nCopyLen);
				m_tBlog.aszTitle[nCopyLen] = '\0';
			}
			else
			{
				strncpy(m_tBlogPosts[nPos]->aszTitle, pszValue, nCopyLen);
				m_tBlogPosts[nPos]->aszTitle[nCopyLen] = '\0';
			}
			break;
		case PTAG_LINK :
			if(m_bHeadSet == false)
			{
				strncpy(m_tBlog.aszLink, pszValue, nCopyLen);
				m_tBlog.aszLink[nCopyLen] = '\0';
			}
			else
			{
				strncpy(m_tBlogPosts[nPos]->aszLink, pszValue, nCopyLen);
				m_tBlogPosts[nPos]->aszLink[nCopyLen] = '\0';
			}
			break;
		case PTAG_DESC :
			if(m_bHeadSet == false)
			{
				strncpy(m_tBlog.aszDesc, pszValue, nCopyLen);
				m_tBlog.aszDesc[nCopyLen] = '\0';
			}
			else
			{
				/* 내용은 새로 메모리 할당 */
				m_tBlogPosts[nPos]->pszDescription = new char[nLen+1];
				if(!m_tBlogPosts[nPos]->pszDescription)
				{
					SetErrorLine(__LINE__);
					return false;
				}
				strcpy(m_tBlogPosts[nPos]->pszDescription, pszValue);
			}
			break;
		case PTAG_DATE :
			if(m_bHeadSet == true)
			{
				SetPostTime(nPos, pszValue);
			}	
			break;
		default :
			SetErrorLine(__LINE__);
			return false;
	}

	return true;
}

void CBlogRss::SetPostTime(int nPos, const char* pszValue)
{
	static char* s_aszMon[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" }; 
	
	int nTemp;
	char aszTemp1[8];
	char aszTemp2[8];
	char aszMonth[8];

	if(nPos < 0 || nPos > m_nPostCount)
		return;

	T_DateTime* pTime = &m_tBlogPosts[nPos]->tTime;

	/* Mon, 19 Nov 2007 21:12:56 +0900 과 같은 형식만 허용 */
	sscanf(pszValue, "%s %d %s %d %d:%d:%d %s", aszTemp1, &pTime->nDay, aszMonth, &pTime->nYear, &pTime->nHour, &pTime->nMin, &nTemp, aszTemp2);
	for(int i = 0; i < 12; i++) 
	{
		if(!strcmp(aszMonth, s_aszMon[i]))
		{
			pTime->nMonth = i + 1;
			break;
		}
	}
}

bool CBlogRss::ParseData(const char* pszData)
{
	char cVal;

	bool bTag, bCdata;

	int nCur, nData;
	char aszTag[128];
	int nCurData = -1;

	bTag =  bCdata = false;
	nCur = nData= 0;

	char* pszBuff = new char[RBUFFER_SIZE];

	unsigned int nLen = strlen(pszData);

	for(unsigned int i = 0; i < nLen; i++)
	{
		cVal = pszData[i];

		if(cVal == '\0')
			break;

		char aa[128];
		if(bTag == true)
		{
			aszTag[nCur++] = cVal; // 태그 정보 저장
		}
		else if(nCurData != -1)
		{
			/* 버퍼에 내용을 저장한다 */
			pszBuff[nData++] = cVal;
			if(bCdata == false && nData == 9)
			{
				memcpy(aa, pszBuff, 9);
				aa[9] = '\0';

				/** CDATA일 경우에는 bCdata 설정 */
				if(!strncmp(aa, "<![CDATA[", 9))
				{
					nData = 0;
					bCdata = true;
				}	
			}

			if(bCdata == true && nData > 4)
			{
				memcpy(aa, pszBuff+nData-4, 3);
				aa[3] = '\0';

				/** CDATA 종료일 경우 */
				if(!strncmp(aa, "]]>", 3))
				{
					nData -= 3;
					bCdata = false;
				}
			}
		}

		/** 태그가 시작될 경우 */
		if(cVal == '<' && bTag == false && bCdata == false)
		{
			memcpy(aa, pszData+i, 9);
			aa[9] = '\0';
			
			if(strncmp(aa, "<![CDATA[", 9))
			{
				bTag = true;
			}	
		}

		/** 태그가 종료될 경우 */
		if(cVal == '>' && bTag == true && bCdata == false)
		{
			aszTag[nCur-1] = '\0';

			if(aszTag[0] == '/')
			{
			/* End(/) 태그일 경우 */
				if(!strcmp(aszTag, "/channel"))
				{
					break;
				}

                for(int k = 0; k < MAX_PTAG-1; k++)
                {
					/* 저장할 태그면 정보를 저장한다 */
                    if(!strcmp(aszTag+1, CBlogRss::m_paszTag[k]))
                    {
						pszBuff[nData-1] = '\0';

						if(strlen(pszBuff) > 0)
							SetPostData(k, pszBuff);

                        nCurData = 0;
						nData = 0;
                        break;
                    }
                }
			}
			else
			{
				bool bOut = false;
				for(int k = 0; k < MAX_PTAG; k++)
				{
					if(!strcmp(aszTag, CBlogRss::m_paszTag[k]))
					{
						if(!strcmp(aszTag, "item"))
						{
							m_bHeadSet = true; // 블로그 정보 종료, 포스트 시작

							/* 새로운 포스트 할당. 포스트 수(MAX_READ_COUNT)를 초과했을 경우에는 루프를 나간다 */
							if(AddPost() == false)
							{
								bOut = true;
								break;
							}
						}	
						nData = 0;
						nCurData = k;	
						break;
					}
				}	
				if(bOut)
					break;
			}

			bTag = false;
			nCur = 0;
		}
	}

	delete pszBuff;

	return true;
}

bool CBlogRss::Connect(const char* pszUrl, const char* pszPath)
{
    int nRt;
    struct hostent* pHost;
    sockaddr_in tServerInfo;

    m_nSocket = socket(AF_INET, SOCK_STREAM, 0);
	if(m_nSocket < 1)
		return false;

    bzero(&tServerInfo, sizeof(sockaddr_in));

    tServerInfo.sin_family = AF_INET;
    tServerInfo.sin_port = htons(80);

    pHost = gethostbyname(pszUrl);
    if(pHost == 0)
	{
		close(m_nSocket);
		m_nSocket = -1;
        return false;
    }    
    tServerInfo.sin_addr.s_addr = *((unsigned int*)(pHost->h_addr_list[0]));
    nRt = connect(m_nSocket, (sockaddr *) &tServerInfo, sizeof(sockaddr_in));
    if(nRt < 0)
    {
		close(m_nSocket);
		m_nSocket = -1;
        return false;
    }

	return true;
}

ssize_t CBlogRss::Read(char* pSou, size_t nMaxLen)
{
    char* pBuff;
    char cData;
    unsigned int nRead, nCount;
    
    if(pSou == NULL)
        return 0;

    if(m_nSocket < 0)
        return -1;

    pBuff = pSou;

    for(nRead = 1; nRead < nMaxLen; nRead++)
    {
AGAIN:
        nCount = read(m_nSocket, &cData, 1);
        if(nCount == 1)
        {
            *pBuff++ = cData;
            if(cData == '\0')
                break;
        }
        else if(nCount == 0)
        {
            if(nRead == 1)
                return 0;
            else
                break;
        }
        else
        {
            if(errno == EINTR)
                goto AGAIN;

            return -1;
        }
    }
    *pBuff = '\0';

    return nRead;
}

ssize_t CBlogRss::Write(const char *pSou, size_t nLength)
{
    size_t nPos;
    ssize_t nCount;
    const char* pBuff;

    pBuff = pSou;
    nPos = nLength;

    if(pSou == NULL)
        return 0;
    
    if(m_nSocket < 0)
        return -1; 

    while(nPos > 0)
    {   
        nCount = write(m_nSocket, pBuff, nPos);
        if(nCount <= 0)
        {
            if(errno == EINTR)
                nCount = 0;
            else
                return -1; 
        }

        nPos -= nCount;
        pBuff += nCount;
    }   
 
    return nCount;
}
